#include "stdafx.h"

namespace WorkbenchLib
{

BOOL GetCommandForFileType (const char *_extension, STRING &command, STRING &err_msg, const char *_action/*=0*/)
{
	STRING extension, action;
	STRING action_key_name;
	HKEY extension_key = 0;
	HKEY action_key = 0;
    char s_action_set_key_name[64];
	MemSet (s_action_set_key_name,0,64);
    long sz = 64;
    char s_command[128];
	MemSet (s_command,0,128);
    long sz_command = 128;

	if (_extension[0] == '.')
		extension = _extension;
	else
	{
		extension = ".";
		extension += _extension;
	}

	if (_action == 0)
		action = "open";
	else
		action = _action;

    long iret = RegOpenKeyEx(HKEY_CLASSES_ROOT,extension,0,KEY_READ,&extension_key);
    if (iret != ERROR_SUCCESS) 
    {
		err_msg = "Extension: ";
		err_msg += (const char *) extension;
		err_msg += " is not registered.";
		iret = FALSE;
		goto cleanup;
    }

	iret = RegQueryValue (extension_key,NULL,s_action_set_key_name,&sz);
    if (iret != ERROR_SUCCESS)
    {
		err_msg = "Problem looking up information for extension: ";
		err_msg += (const char *) extension;
		err_msg += " Windows registry may be corrupt.";
		iret = FALSE;
		goto cleanup;
    }

	iret = RegOpenKeyEx(HKEY_CLASSES_ROOT,s_action_set_key_name,0,KEY_READ,&action_key);
	if (iret != ERROR_SUCCESS)
    {
		err_msg = "Problem looking up information for extension: ";
		err_msg += (const char *) extension;
		err_msg += " Windows registry may be corrupt.";
		iret = FALSE;
		goto cleanup;
    }

	iret = RegOpenKeyEx(action_key,"shell",0,KEY_READ,&action_key);
	if (iret != ERROR_SUCCESS)
    {
		err_msg = "Problem looking up information for extension: ";
		err_msg += (const char *) extension;
		err_msg += " Windows registry may be corrupt.";
		iret = FALSE;
		goto cleanup;
    }

	iret = RegOpenKeyEx(action_key,action,0,KEY_READ,&action_key);
	if (iret != ERROR_SUCCESS)
 	{
		err_msg = "Extension: ";
		err_msg += (const char *) extension;
		err_msg += " is registered.  But does not have action: ";
		err_msg += (const char *) action;
		iret = FALSE;
		goto cleanup;
	}

	iret = RegOpenKeyEx(action_key,"command",0,KEY_READ,&action_key);
	if (iret != ERROR_SUCCESS)
	{
		err_msg = "Problem looking up information for extension: ";
		err_msg += (const char *) extension;
		err_msg += " Windows registry may be corrupt.";
		iret = FALSE;
		goto cleanup;
    }

	iret = RegQueryValue (action_key,NULL,s_command,&sz_command);
    if (iret != ERROR_SUCCESS && iret != ERROR_INVALID_DATA)
    {
		err_msg = "Problem looking up information for extension: ";
		err_msg += (const char *) extension;
		err_msg += " Windows registry may be corrupt.";
		iret = FALSE;
		goto cleanup;
    }

	command = s_command;
	iret = TRUE;

cleanup:

	if (action_key) RegCloseKey (action_key);
	if (extension_key) RegCloseKey (extension_key);

    return iret;
}

MCTRL_EXPORT BOOL OpenFileUsingDefaultApp (const char *file_name, const char *foreign_ext/*=0*/)
{
	// Clean up the file name
	STRING _file_name = file_name;
	int n = _file_name.Replace("\\\\", "\\");
	n += _file_name.Replace("/", "\\");

	if (n)
		file_name = _file_name;

	BYTE v_return = TRUE;

	while (1)
	{
		STRING extension, command, err_msg;

		if (foreign_ext == 0)
			extension.LoadFileName(file_name, FALSE, FALSE, TRUE);
		else
			extension = foreign_ext;

		if (GetCommandForFileType(extension, command, err_msg, 0) == FALSE)
		{
			if (foreign_ext)
			{
				extension.LoadFileName(file_name, FALSE, FALSE, TRUE);
				if (GetCommandForFileType(extension, command, err_msg, 0) == FALSE)
				{
					v_return = FALSE;
					break;
				}
			}
			else
			{
				v_return = FALSE;
				break;
			}
		}

		/////////////////////////////////////////////////////////////////////////////////////
		// Get ready to spawn the process for the compiler.  Inside of this process the
		// compiler will compile one source code file.
		//
		BOOL                bInheritHandles;
		DWORD                dwCreationFlags;
		STARTUPINFO            StartupInfo;
    
		PROCESS_INFORMATION ProcessInfo;
    
		// This parameter must be true in order to pipe the stdout and stderr streams
		bInheritHandles = FALSE;
		dwCreationFlags = 0;
		MemSet (&StartupInfo, 0, sizeof(StartupInfo));
		MemSet (&ProcessInfo, 0, sizeof(ProcessInfo));

		StartupInfo.dwFlags = STARTF_USESHOWWINDOW;
		StartupInfo.wShowWindow = SW_RESTORE;

		STRING command_line;

		const char *c = ::strstr (command,"%1");
		if (c)
		{
			const char *c0 = command;
			int idx = (int)(c-c0);

			command.Remove (idx,2);
			STRING *tmp = command.Cut (idx);
			command += "\"";
			command += (const char *) file_name;
			command += "\"";
			command += (const char *) (*tmp);
			delete tmp;
			command_line = (const char *) command;
		}
		else
		{
			command_line = (const char *) command;
			command_line += " \"";
			command_line += file_name;
			command_line += "\"";
		}

		c = ::strstr (command_line,"%SystemRoot%");
		if (c)
		{
			const char *c0 = command_line;
			int idx = (int)(c-c0);

			command_line.Remove (idx,13);

			char system_dir[128];

			GetSystemDirectory(system_dir,128);
			STRING path;
			path = system_dir;
			path.RemoveSubDir ();
			path += "\\";


			command_line.Insert (idx,path,StrLen (path));
		}
    
		char _command_line[1024];
		strncpy_s (_command_line,sizeof(_command_line),(const char *) command_line,1024);
    
		if(!CreateProcess(           NULL,                    // Application name
								   _command_line,            // Command line
								   NULL,                    // Process Attributes
								   NULL,                    // Thread Attributes
								   bInheritHandles,            // Inherit Handles (TRUE,FALSE)
								   dwCreationFlags,            // Creation flags
								   NULL,                    // environment block - use Create's
								   "c:\\",                        // current directory of child process - use Create's
								   &StartupInfo,            
								   &ProcessInfo))
		{
			// CreateProcess failed - find out why and tell user
			DWORD dwLastError;
			dwLastError = GetLastError();
			v_return = FALSE;
		}

		break;
    }

	return v_return;
}

};