#include "stdafx.h"

namespace WorkbenchLib
{

BOOL SetRootFolder(LPCTSTR lpszRootFolder, BROWSEINFO &bi)
{
	LPITEMIDLIST  pidl;
	LPSHELLFOLDER pDesktopFolder;
	OLECHAR       olePath[MAX_PATH];
	ULONG         chEaten;
	ULONG         dwAttributes;
	HRESULT       hr;

	// Get a pointer to the Desktop's IShellFolder interface.
	if (SUCCEEDED(SHGetDesktopFolder(&pDesktopFolder)))
	{

		// IShellFolder::ParseDisplayName requires the file name be in
		// Unicode.
		MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, lpszRootFolder, -1,
                             olePath, MAX_PATH);

         // Convert the path to an ITEMIDLIST.
         hr = pDesktopFolder->ParseDisplayName(NULL,
                                               NULL,
                                               olePath,
                                               &chEaten,
                                               &pidl,
                                               &dwAttributes);
         if (FAILED(hr))
         {
                     bi.pidlRoot = NULL;
             return FALSE;// Handle error.
         }

         bi.pidlRoot = pidl;
         //release the desktop folder object
         pDesktopFolder->Release();
         return TRUE;
     }

     bi.pidlRoot = NULL;
     return FALSE;
}

int CALLBACK SetSelProc (HWND hWnd,
                         UINT uMsg,
                         LPARAM lParam,
                         LPARAM lpData)
{
   if (uMsg==BFFM_INITIALIZED)
   {
      ::SendMessage(hWnd, BFFM_SETSELECTION, TRUE, lpData );
   }
   return 0;
}

MCTRL_EXPORT BOOL SelectDirectory (STRING &dir, const char *title, const char *root_folder/*=0*/)
{
	BROWSEINFO info;
	MemSet (&info,0,sizeof(info));
	char path[MAX_PATH];
	info.pszDisplayName = path;

	info.lpszTitle = title;

	info.ulFlags = BIF_EDITBOX;
//	info.ulFlags |= BIF_DONTGOBELOWDOMAIN;
	
	if (root_folder)
		SetRootFolder (root_folder,info);

	if (dir.HasLen())
	{
		// Set the callback Fn pointer to my function:
		info.lpfn   = SetSelProc;
		// Set the lParam to point at my buffer containing the folder:
		info.lParam = (LPARAM)(const char *)dir;
	}
	else
	{
		info.lpfn   = NULL;
		info.lParam = 0;
	}

	ITEMIDLIST *location = ::SHBrowseForFolder(&info);

	if (location)
	{ 
		SHGetPathFromIDList(location, path);
		dir = path;

		// Free the location object
		IMalloc *imalloc;
		if (SHGetMalloc (&imalloc) == S_OK)
		{
			if (imalloc)
			{
				imalloc->Free(location);
				imalloc->Release ();
			}
		}

		return TRUE;
	}

	return FALSE;
}

};